
#include <math.h>
#include <stdlib.h> 

#include "vlabvolt.h"

using namespace std;

GPoint& GPoint::normalize(void)
    { *this /= sqrt(X[0]*X[0] + X[1]*X[1] + X[2]*X[2]);  return *this; }

GPoint GPoint::cross(const GPoint& p) const              // return *this cross p
{
  return GPoint(X[1] * p.X[2] - X[2] * p.X[1],
               X[2] * p.X[0] - X[0] * p.X[2],
               X[0] * p.X[1] - X[1] * p.X[0]);
}

double GPoint::dot(const GPoint& p) const 
  { return X[0]*p.X[0] + X[1]*p.X[1] + X[2]*p.X[2]; }

istream& operator>> (istream& s, GPoint& p)
{
  s >> p.X[0] >> p.X[1] >> p.X[2];
  if (!s) {cout<<("Error reading GPoint from stream\n\n"); exit(1);}
  return s;
}

ostream& operator<< (ostream& s, const GPoint& p)
{
  s << p.X[0] << "   " << p.X[1] << "   " << p.X[2];
  if (!s) {cout<<("Error writing GPoint to stream"); exit(1);}
  return s;
}

Transform::Transform(void)	{
	int i, j;
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			val[i][j] = (i==j);
		}
	}
}

Transform::Transform(const double q[4], const GPoint &t) {
	val[0][0] = q[0]*q[0] + q[1]*q[1] - q[2]*q[2] - q[3]*q[3];
	val[0][1] = 2. * (q[1]*q[2] - q[0]*q[3]);
	val[0][2] = 2. * (q[1]*q[3] + q[0]*q[2]);
	val[0][3] = t[0];
	val[1][0] = 2. * (q[1]*q[2] + q[0]*q[3]);
	val[1][1] = q[0]*q[0] - q[1]*q[1] + q[2]*q[2] - q[3]*q[3];
	val[1][2] = 2. * (q[2]*q[3] - q[0]*q[1]);
	val[1][3] = t[1];
	val[2][0] = 2. * (q[1]*q[3] - q[0]*q[2]);
	val[2][1] = 2. * (q[2]*q[3] + q[0]*q[1]);
	val[2][2] = q[0]*q[0] - q[1]*q[1] - q[2]*q[2] + q[3]*q[3];
	val[2][3] = t[2];
	val[3][0] = val[3][1] = val[3][2] = 0.;
	val[3][3] = 1.;
}


Transform::Transform(double d, double theta, double a, double alpha) {
	double ct = cos(theta), st = sin(theta), ca = cos(alpha), sa = sin(alpha);
	val[0][0] = ct;
	val[0][1] = -ca * st;
	val[0][2] = sa * st;
	val[0][3] = a * ct;
	val[1][0] = st;
	val[1][1] = ca * ct;
	val[1][2] = -sa * ct;
	val[1][3] = a * st;
	val[2][0] = 0.;
	val[2][1] = sa;
	val[2][2] = ca;
	val[2][3] = d;
	val[3][0] = val[3][1] = val[3][2] = 0.;
	val[3][3] = 1.;
}

Transform Transform::inverse(void) const {
	Transform result;
	int i,j;

	for (i=0; i<3; i++) {
		result.val[i][3] = 0;
		for (j=0; j<3; j++) {
			result.val[i][j] = val[j][i];
			result.val[i][3] -= val[j][i] * val[j][3];
		}
	}

	return result;
}

Transform Transform::operator*(const Transform& t) const { 
	Transform result;
	int i, j, k;

	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			result.val[i][j] = 0;
			for (k=0; k<4; k++) result.val[i][j] += val[i][k] * t.val[k][j];
		}
	}
	return result; 
}

GPoint Transform::operator*(const GPoint& p) const { 
	GPoint result;
	int i, j;

	for (i=0; i<3; i++) {
		result[i] = val[i][3];
		for (j=0; j<3; j++) result[i] += val[i][j] * p[j];
	}
	return result; 
}

GPoint Transform::rotatedPoint(const GPoint& p) const {
	GPoint result;
	int i, j;

	for (i=0; i<3; i++) {
		result[i] = 0.;
		for (j=0; j<3; j++) result[i] += val[i][j] * p[j];
	}
	return result; 
}


istream& operator>>(istream& s, Transform& p)
{
	int i, j;
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			s >> p.val[i][j];
		}
	}
	if (!s) {cout<<("Error reading GPoint from stream\n\n"); exit(1);}
	return s;
}


ostream& operator<<(ostream& s, const Transform& p)
{
	int i, j;
	for (i=0; i<4; i++) {
		for (j=0; j<4; j++) {
			s << setprecision(6) << p.val[i][j] << "  ";
		}
		s << endl;
	}
	s << endl;

	if (!s) {cout<<("Error writing GPoint to stream"); exit(1);}
	return s;
}

void CallBackFunc(int event, int x, int y, int flags, void* userdata)
{
     if  ( event == EVENT_LBUTTONDOWN )
     {
          cout << "Left button of the mouse is clicked - position (" << x << ", " << y << ")" << endl;
     }
     else if  ( event == EVENT_RBUTTONDOWN )
     {
          cout << "Right button of the mouse is clicked - position (" << x << ", " << y << ")" << endl;
     }
     else if  ( event == EVENT_MBUTTONDOWN )
     {
          cout << "Middle button of the mouse is clicked - position (" << x << ", " << y << ")" << endl;
     }
     else if ( event == EVENT_MOUSEMOVE )
     {
          cout << "Mouse move over the window - position (" << x << ", " << y << ")" << endl;
     }
}


Viewport::Viewport(int w, const char* wname) : width(w), windowname(wname), backgroundcolor(255,140,140), 
	canvas(width, width, CV_8UC3, backgroundcolor) {

#if 0
	 Mat tmp(400,400, CV_8UC3, Scalar(128,128,128));
     //Create a window
     namedWindow("My Window", 1);

     //set the callback function for any mouse event
     setMouseCallback("My Window", CallBackFunc, NULL);

     //show the image
     imshow("My Window", tmp);
#endif

     namedWindow(windowname, 1);
//     setMouseCallback(windowname, CallBackFunc, NULL);
		imshow( windowname, canvas );
		moveWindow( windowname, 200, 100 );		// move upper left corner of window to col, row

//		setMouseCallback(windowname, CallBackFunc, NULL);

}

void Viewport::refresh(void) {
	imshow( windowname, canvas );
	canvas.release();										// Create new canvas for drawing
	canvas = Mat(width, width, CV_8UC3, backgroundcolor); 
}

void Viewport::drawline(Point start, Point end, const Scalar & color, int thickness, int lineType) const {
	line(canvas, start, end, color, thickness, lineType); 
}

inline void Viewport::drawpoly(const Point * pts, int npts, const Scalar & color, int lineType, int shift) {
	fillConvexPoly(canvas, pts, npts, color, lineType, shift);
}

// method 2 --- This one allows non-convex shapes
inline void Viewport::drawpoly2(const Point * pts, int npts, const Scalar & color, int lineType, int shift) {
	fillPoly(canvas, &pts, &npts, 1, color, lineType, shift);
}

void Viewport::drawpolyboundary(const Point * pts, int npts, const Scalar & color, int thickness, int lineType) const {
	int i;
	for (i=1; i<npts; i++) {
		drawline(pts[i-1], pts[i], color, thickness, lineType);
	}
	drawline(pts[npts-1], pts[0], color, thickness, lineType);
}

inline void Viewport::puttext(const string& text, Point org, int fontFace, double fontScale, const Scalar & color, int thickness, 
	int lineType, bool bottomLeftOrigin) { 

	putText(canvas, text, org, fontFace, fontScale, color, thickness, lineType, bottomLeftOrigin) ;
}


void Camera::setTransform(void) { 
	Transform wTc;		// The camera wrt world
	int i;

	GPoint zVec(-GPoint(viewpoint).normalize());		// Negative of normalized version of viewpoint
	GPoint xVec(GPoint(-zVec[1], zVec[0], 0).normalize());	// Normal to Z and horizontal
	GPoint yVec(zVec.cross(xVec));		// Right hand rule

	for (i=0; i<3; i++) {
		wTc.setval(i,0, xVec[i]); 
		wTc.setval(i,1, yVec[i]); 
		wTc.setval(i,2, zVec[i]); 
		wTc.setval(i,3, viewpoint[i]); 
	}
	vptrans = wTc.inverse();
}
Point Camera::toImage(const GPoint& p) const {		// Only x and y of result are meaningful
	GPoint pcam(vptrans * p);		// GPoint in camera XYZ
	GPoint imagexy = pcam * (scale * (viewport.vpwidth()/2) / (1 - pcam[2])) + centershift;	// convert to image coordinates in cm
																		// Assign lambda = 1
//	cout << "Point: " << p << "  pcam: " << pcam << "  imagexy:  " << imagexy ;
//	cout << "  ix:  " << (viewport.vpwidth()/2 + (int)imagexy[0]) 
//     << "  iy:  " << (viewport.vpwidth()/2 + (int)imagexy[1]) << endl;	// convert to image coordinates
	return Point( viewport.vpwidth()/2 + (int)imagexy[0], viewport.vpwidth()/2 + (int)imagexy[1] );	// convert to image coordinates
}

void Camera::shiftviewpoint(camdirection direction) {
	static const double lrQuaternion[] = {cos(deg2rad(10./2)), 0., 0., sin(deg2rad(10./2))};
	Transform lrTrans(lrQuaternion, GPoint());
	GPoint viewdirXY(viewpoint[0], viewpoint[1], 0); 		// XY components of viewdirection
	viewdirXY.normalize();
	const double udQuaternion[] = {cos(deg2rad(10./2)), viewdirXY[1]*sin(deg2rad(10./2)), -viewdirXY[0]*sin(deg2rad(10./2)), 0};
	Transform udTrans(udQuaternion, GPoint());
	GPoint viewdir(viewpoint);
	viewdir.normalize();
	double viewdist2 = viewpoint.dot(viewpoint);		// view distance squared
	switch (direction) {
		case camleft:
			if (viewdirXY[0] > 0.2 || viewdirXY[1] > 0.) viewpoint = lrTrans.inverse() * viewpoint;
			break;
		case camright:
			if (viewdirXY[0] > 0.2 || viewdirXY[1] < 0.) viewpoint = lrTrans * viewpoint;
			break;
		case camup:
			if (viewdir[2] < 0.95) viewpoint = udTrans * viewpoint;
			break;
		case camdown:
			if (viewdir[2] > 0.11) viewpoint = udTrans.inverse() * viewpoint;
			break;
		case camin:
			if (viewdist2 > 100*100) viewpoint -= 20*viewdir;
			break;
		case camout:
			if (viewdist2 < 2000*2000) viewpoint += 20*viewdir;
			break;


	}

	setTransform();

} 

void Camera::drawbase(void) const {
	for (int y=-50; y<=50; y+=10) drawline(GPoint(0,y,0), GPoint(50,y,0), BLACK, 1);	//Grid is line thickness of 1
	for (int x=0; x<=50; x+=10) drawline(GPoint(x,-50,0), GPoint(x,50,0), BLACK, 1);
	drawline(GPoint(0,0,0), GPoint(100,0,0), RED);
	drawline(GPoint(0,0,0), GPoint(0,100,0), GREEN);
	drawline(GPoint(0,0,0), GPoint(0,0,100), BLUE);
}

void Camera::drawline(GPoint start, GPoint end, const Scalar & color, int thickness, int lineType) const {
	viewport.drawline(toImage(start), toImage(end), color, thickness, lineType);
}

void Camera::drawpoly(const vector<GPoint> & gpts, GPoint surfnorm, const Scalar & color, bool isconvex, int lineType, int shift) {
	int npts = (int)gpts.size();

// This check tests if camera viewpoint (translated to object frame) is "above the plane"
	if ((vptrans.inverse() * GPoint() - gpts[0]).dot(surfnorm) < 0) return;	// Skip drawing if surface norm points away from camera
//	if ((vptrans * gpts[0]).dot(vptrans * surfnorm - vptrans * GPoint()) > 0) return;	// Equivalent
//	cout << "Norm: " << surfnorm << endl;

	vector<Point> pts(npts);
	vector<GPoint>::const_iterator gpit = gpts.begin();
	for (vector<Point>::iterator pit = pts.begin(); pit != pts.end(); pit++) {
		*pit = toImage(*gpit++);
	}

	int grayvalue = int(127*surfnorm.dot(lightsource)) + 128;
	Scalar paintcolor;
	if (color==RED) paintcolor = Scalar(0, 0, grayvalue);
	else if (color==GREEN) paintcolor = Scalar(0, grayvalue, 0);
	else if (color==BLUE) paintcolor = Scalar(grayvalue, 0, 0);
	else paintcolor = Scalar(grayvalue, grayvalue, grayvalue);
	if (isconvex) viewport.drawpoly(&pts[0], npts, paintcolor, lineType, shift) ;
	else viewport.drawpoly2(&pts[0], npts, paintcolor, lineType, shift) ;
	viewport.drawpolyboundary(&pts[0], npts, BLACK, 1, lineType) ;
}

void Camera::puttext(const string& text, int linenum, int fontFace, double fontScale, const Scalar & color, int thickness, 
	int lineType, bool bottomLeftOrigin) { 

	viewport.puttext(text, Point(5,15*linenum), fontFace, fontScale, color, thickness, lineType, bottomLeftOrigin) ;
}


void Frame::draw(Camera & cam) const {
	const double length = 20;
	GPoint origin(val[0][3], val[1][3], val[2][3]);
	cam.drawline(origin, origin + length * GPoint(val[0][0], val[1][0], val[2][0]), RED);
	cam.drawline(origin, origin + length * GPoint(val[0][1], val[1][1], val[2][1]), GREEN);
	cam.drawline(origin, origin + length * GPoint(val[0][2], val[1][2], val[2][2]), BLUE);
//	cam.drawline(origin, origin - GPoint(0., 0., val[2][3]), GRAY);
}


void Surface::resurface(const vector<GPoint> & pts, const vector<int> & indices, bool iscvx) {
	points.resize(indices.size());
	isconvex = iscvx;
	vector<GPoint>::iterator pit = points.begin();
	for (vector<int>::const_iterator it = indices.begin(); it != indices.end(); it++) {
		*pit++ = pts[*it];
	}
	setnormal();
} 

void Surface::resurface4(const GPoint & p0, const GPoint & p1, const GPoint & p2, const GPoint & p3)  {
	points.resize(4);
	points[0] = p0;
	points[1] = p1;
	points[2] = p2;
	points[3] = p3;
	setnormal();
} 

void Surface::draw(Camera & cam, Transform t, const Scalar & color, int lineType) const {
	vector<GPoint> transpts(points.size());		// Array of the points transformed by "t"

	vector<GPoint>::iterator trans_it = transpts.begin();		// transform all points
	for (vector<GPoint>::const_iterator it = points.begin(); it != points.end(); it++) {
		*trans_it++ = t * *it;
	}
	
	cam.drawpoly(transpts, t.rotatedPoint(normal), color, isconvex, lineType);
}


Object::Object(stringstream & strobj) {
	int numpoints, numsurfaces, numvertices, isconvex;

	strobj >> numpoints;
	vector<GPoint> points(numpoints);
	vector<GPoint>::iterator pit = points.begin();
	for (vector<GPoint>::iterator it = points.begin(); it != points.end(); it++) {
		strobj >> (*it);
	}

	strobj >> numsurfaces;
	surfaces.resize(numsurfaces);
	for (vector<Surface>::iterator surfit = surfaces.begin(); surfit != surfaces.end(); surfit++) {
		strobj >> numvertices >> isconvex;
		vector<int> indx(numvertices);
		for (vector<int>::iterator it = indx.begin(); it != indx.end(); it++) {
			strobj >> (*it);
		}
		surfit->resurface(points, indx, (bool)isconvex); 
	}

	if (!strobj) {cout<<("Error reading Object from stream\n\n"); exit(1);}
}

void Object::draw(Camera & cam, const Scalar & color, int lineType) const {
	for (vector<Surface>::const_iterator it = surfaces.begin(); it != surfaces.end(); it++) {
		it->draw(cam, trans, color, lineType);
	}
//	Frame(trans).draw(cam);
}


Block::Block(GPoint d2, Transform t) : Object(t), dims2(d2) {
	double x = d2[0];
	double y = d2[1];
	double z = d2[2];

	GPoint pts[8] = {GPoint(-x,-y,0.), GPoint(x,-y,0.), GPoint(-x,y,0.), GPoint(x,y,0.), 
					 GPoint(-x,-y,2.*z),  GPoint(x,-y,2.*z),  GPoint(-x,y,2.*z),  GPoint(x,y,2.*z)};

	surfaces[0].resurface4(pts[4], pts[5], pts[7], pts[6]);	// Top is first
	surfaces[1].resurface4(pts[0], pts[1], pts[5], pts[4]);
	surfaces[2].resurface4(pts[1], pts[3], pts[7], pts[5]);
	surfaces[3].resurface4(pts[3], pts[2], pts[6], pts[7]);
	surfaces[4].resurface4(pts[2], pts[0], pts[4], pts[6]);
	surfaces[5].resurface4(pts[0], pts[2], pts[3], pts[1]);	// Bottom is last
}


void Robot::convertsteps(void) {		// Sets Robot::theta based on the stepsfromhome
	int *sfh = stepsfromhome;			// Simply a shorthand for stepsfromhome
	theta[0] = deg2rad((double)sfh[0] * 360. / 6500.);
	theta[1] = deg2rad((double)sfh[1] * 360. / 8600.);
	theta[2] = deg2rad((double)(sfh[1] + sfh[2]) * 360. / 8600.);
	theta[3] = deg2rad(((double)(-sfh[2]) / 8600 + (double)(sfh[3] - sfh[4]) / 13000) * 360);
	theta[4] = -deg2rad(((double)(sfh[3] + sfh[4]) / 13000) * 360);
}

void Robot::setTransforms(void) {
	const double q[4] = { 1., 0., 0., 0.};
	const Transform trOpen(q, GPoint(0., 3.0, 0.9));
	const Transform trOpenInv(q, GPoint(0., -3.0, 0.9));
	const Transform trClosed(q, GPoint(0., 1.5, 0.9));
	const Transform trClosedInv(q, GPoint(0., -1.5, 0.9));

	for (int i = 0; i < 5; i++) {
		Transform trelative(d[i], theta[i]+thetaoffset[i], a[i], alpha[i]);	// relative to previous
		links[i+1].setTransform(links[i].getTransform() * trelative);
	}

	if (gripperClosed) {
		links[6].setTransform(links[5].getTransform() * trClosed);
		links[7].setTransform(links[5].getTransform() * trClosedInv);
		if (attachedblock != NULL) attachedblock->setTransform(links[5].getTransform() * blockreltrans); 
		
	}
	else {
		links[6].setTransform(links[5].getTransform() * trOpen);
		links[7].setTransform(links[5].getTransform() * trOpenInv);
	}

	// Note the fingers, when open,  (as well as link0) are the same as the "previous"
}

void Robot::checklimits(void) {			// Return true if all axes within limits and z above table
	if (
		theta[0] < deg2rad(-150.) || theta[0] > deg2rad(150.) || 
		theta[1] < deg2rad(-30.) || theta[1] > deg2rad(150.) ||
		theta[2] < deg2rad(-20.) || theta[2] > deg2rad(150.) ||
		theta[3] < deg2rad(-100.) || theta[3] > deg2rad(100.) ||
		theta[4] < deg2rad(-360.) || theta[4] > deg2rad(360.) ||
		links[5].getTransform().gettranslation()[2] < 0.5
	) exceededlimits =  true;
}

void Robot::putinfo(Camera & cam) {				// put information on screen
	stringstream thetastring;
	thetastring << fixed << setprecision(1) << rad2deg(theta[0]) << "  " << rad2deg(theta[1]) << "  " << rad2deg(theta[2]) << "  " << rad2deg(theta[3]) << "  " << rad2deg(theta[4]);
	stringstream posstring;
	posstring << fixed << setprecision(1) << links[5].getTransform().gettranslation();
	cam.puttext(thetastring.str(), 1);
	cam.puttext(posstring.str(), 2);
	if (exceededlimits) cam.puttext("ROBOT EXCEEDED LIMITS", 3, FONT_HERSHEY_PLAIN, 1, RED, 2);
}


void Robot::draw(Camera & cam, int lineType) {
	for (vector<Object>::iterator it = links.begin(); it != links.end(); it++) {
		it->draw(cam, WHITE, lineType);
	}
//	int *sfh = stepsfromhome;
//	cout << sfh[0] << "  " << sfh[1] << "  " << sfh[2] << "  " << sfh[3] << "  " << sfh[4] << endl;
}

void Robot::redrawfinger(Camera & cam, int lineType) {	// redraw the finger which could have been covered by a block
	if ((links[5].getTransform().inverse() * cam.getviewpoint()) [1] > 0) {
		links[6].draw(cam, WHITE, lineType);
	} 
	else {
		links[7].draw(cam, WHITE, lineType);
	}
}

bool Robot::moverel(int steps[], bool fullmove) {	// return steps[] with steps remaining 
	int target[5];

	for (int i=0; i<5; i++) {
		target[i] = stepsfromhome[i] + steps[i];
	}

	bool arrivedtarget = true;
	for (int i=0; i<5; i++) {
		if (fullmove) {
			stepsfromhome[i] = target[i];
		}
		else if (stepsfromhome[i] < target[i]) {
			stepsfromhome[i] += stepsize[i];
			if (stepsfromhome[i] > target[i]) stepsfromhome[i] = target[i];
		}
		else {
			stepsfromhome[i] -= stepsize[i];
			if (stepsfromhome[i] < target[i]) stepsfromhome[i] = target[i];
		}
		steps[i] = target[i] - stepsfromhome[i];		// Adjust steps[] to steps remaining
		if (steps[i]) arrivedtarget = false;
	}
	convertsteps();
	setTransforms();

	return arrivedtarget;		// true if we arrived at the target
}

const vector<int> Robot::getnestrelative(void) { 		// return moverel steps to get to nest
	vector<int> result(5);
	for (int i=0; i<5; i++) result[i] = nestloc[i] - stepsfromhome[i];
	return result; 
}

void Robot::gripperOpen(Block * blocks, int *ontop, int nblocks) { 
	gripperClosed = false; 
	if (nblocks == 0 || attachedblock==NULL) return;
	Transform btrans = attachedblock->getTransform();	// Transform block laying flat on the table - or block below
// Note trig identies: cos(t/2) = sqrt((1+cos(t))/2) and sin(t/2) = sqrt((1-cos(t))/2)
// cos(theta) = dot product of Zobj and Zworld = Zobj[2]
	double cost = btrans.getval(2,2);
	if (cost > 1.) cost = 1.;
	else if (cost < -1.) cost = -1.;
	double ct2 = sqrt((1.+cost)/2.);			// Cos and sin of angle to rotate / 2
	double st2 = sqrt((1.-cost)/2.);
	GPoint rotaxis(GPoint(btrans.getval(1,2), -btrans.getval(0,2),0).normalize()); // Axis to rotate flat
	double q[4] = {ct2, st2*rotaxis[0], st2*rotaxis[1], st2*rotaxis[2]};	// Rotation as quaternion
	GPoint trans(btrans.gettranslation());

	btrans = Transform(q,GPoint()) * btrans;
	double maxheight = 0.; 							// First assume the table is the height
	int upperblock = -1;
	int thisblock = -1;
	for (int i=0; i<::numblocks; i++) {
		if (&blocks[i] == attachedblock) {		// Don't check this block if it is the one being released
			thisblock = i;
			continue;
		}
		GPoint thiswrtthat( blocks[i].getTransform().inverse() * trans );
		if (thiswrtthat[0] > -blocks[i].getdims2()[0] && thiswrtthat[0] < blocks[i].getdims2()[0] &&
			thiswrtthat[1] > -blocks[i].getdims2()[1] && thiswrtthat[1] < blocks[i].getdims2()[1] ) {
				double thistop = blocks[i].getTransform().gettranslation()[2] + 2.;
				if (thistop > maxheight) {
					maxheight = thistop;
					upperblock = i;					// find index of uppermost block
				}
		}

	}
	if (upperblock != -1) ontop[upperblock] = thisblock;	// Found at least one block so set its ontop
	trans[2] = maxheight;						// Set to table height or height of highest block underneath
	btrans.settranslation(trans);
	attachedblock->setTransform(btrans);		// Do the drop	
	attachedblock = NULL;					// Detach block from the gripper
//	for (int i=0; i<::numblocks; i++) {
//		cout << ontop[i] << " " ;
//	}
//	cout << endl;
}

void Robot::gripperClose(Block * blocks, int *ontop, int nblocks) { 
	if (gripperClosed) return;
	gripperClosed = true; 				// Do this even if ::numblocks is 0
	if (nblocks == 0) return;
	GPoint fingerL(links[6].getTransform().gettranslation());
	GPoint fingerR(links[7].getTransform().gettranslation());
	int blocknum;
	for (blocknum=0; blocknum<::numblocks; blocknum++) {
		GPoint blocktrans(blocks[blocknum].getTransform().gettranslation());
		if ( fingerL[2] > (blocktrans[2] + 2.*blocks[blocknum].getdims2()[2]) ||
		     fingerR[2] > (blocktrans[2] + 2.*blocks[blocknum].getdims2()[2]) ||
		     fingerL[2] < blocktrans[2]  ||
		     fingerR[2] < blocktrans[2]  
		) continue;							// Found one we can grasp
		Transform blockinverse(blocks[blocknum].getTransform().inverse());
		GPoint dims2 = blocks[blocknum].getdims2();
		GPoint fingL = blockinverse * fingerL;		// Fingers in block coordinate frame
		GPoint fingR = blockinverse * fingerR;
		if (fingL[1] < fingR[1]) {			// If wrist rotated 180 degrees
			GPoint tmp(fingL);
			fingL = fingR;
			fingR = tmp;
		}
		if (fingL[1] > dims2[1] && fingL[0] < dims2[0] && fingL[0] > -dims2[0] && 
			fingR[1] < -dims2[1] && fingR[0] < dims2[0] && fingR[0] > -dims2[0] 
		) break;			// Success 
	}
	if (blocknum == ::numblocks || ontop[blocknum] != -1) {		// return if none grasped (or one on top of it)
		attachedblock = NULL;		// Shouldn't be needed
		cout << "Failed to grasp a block" << endl;
		return ;
	}
//	blocknum = 2;
// Now "straighten" the block within the gripper by making the "Y" axes parallel (Y is normal to the gripper surface)
	blockreltrans = links[5].getTransform().inverse() * blocks[blocknum].getTransform();    // rTb = 0Tr.inv * 0Tb
// Note trig identies: cos(t/2) = sqrt((1+cos(t))/2) and sin(t/2) = sqrt((1-cos(t))/2)
// cos(theta) = dot product of Yobj and YGripper = Yobj[1]
	double cost = blockreltrans.getval(1,1);
	if (cost > 1.) cost = 1.;
	else if (cost < -1.) cost = -1.;
	double ct2 = sqrt((1.+cost)/2.);		// Cos and sin of angle to rotate / 2
	double st2 = sqrt((1.-cost)/2.);
	GPoint rotaxis(GPoint(-blockreltrans.getval(2,1), 0., blockreltrans.getval(0,1)).normalize()); // Axis to rotate wrt [5]
	double q[4] = {ct2, st2*rotaxis[0], st2*rotaxis[1], st2*rotaxis[2]};	// Rotation as quaternion
	GPoint trans(blockreltrans.gettranslation());
	trans[1] = 0.;									// centerline of gripper
	blockreltrans = Transform(q,GPoint()) * blockreltrans;		// This gets the rotation part (ignore result translation)
	blockreltrans.settranslation(trans);					// Shift to centerline of gripper
//	blockreltrans.setval(1,3,0.);		// Before
	attachedblock = &blocks[blocknum];
	for (int i=0; i<::numblocks; i++) {		// Find the block with that one on top and clear it's ontop
		if (ontop[i] == blocknum) ontop[i] = -1;
	}
}

stringstream link0string("\
20 \
10.86485 18.415 0. \
-21.83765 18.415 0. \
-21.83765 18.415 1.27 \
10.86485  18.415 1.27 \
3.4036 8.5725 1.27 \
-14.3764 8.5725 1.27 \
-14.3764 8.5725 8.89 \
3.4036 8.5725 8.89 \
-14.3764 4.1275 13.97 \
3.4036 4.1275 13.97 \
10.86485 -18.415 0. \
-21.83765 -18.415 0. \
-21.83765 -18.415 1.27 \
10.86485  -18.415 1.27 \
3.4036 -8.5725 1.27 \
-14.3764 -8.5725 1.27 \
-14.3764 -8.5725 8.89 \
3.4036 -8.5725 8.89 \
-14.3764 -4.1275 13.97 \
3.4036 -4.1275 13.97 \
\
13 \
4 1   0 1 2 3 \
4 1   13 12 11 10 \
4 1   10 0 3 13 \
4 1   1 11 12 2 \
4 1   3 2 12 13 \
4 1   4 5 6 7 \
4 1   17 16 15 14 \
4 1   7 6 8 9 \
4 1   19 18 16 17 \
6 1   14 4 7 9 19 17 \
6 1   5 15 16 18 8 6 \
4 1   9 8 18 19 \
4 1   1 0 10 11 \
");

stringstream link1string("\
12 \
4.445 -13.0 -5.23875 \
-14.3764 -13.0 -5.23875 \
-14.3764 -4.7625 -5.23875 \
-1.841179 4.445 -5.23875 \
1.841179 4.445 -5.23875 \
4.445 1.841179 -5.23875 \
4.445 -13.0 5.23875 \
-14.3764 -13.0 5.23875 \
-14.3764 -4.7625 5.23875 \
-1.841179 4.445 5.23875 \
1.841179 4.445 5.23875 \
4.445 1.841179 5.23875 \
\
8 \
4 1   0 6 7 1 \
4 1   1 7 8 2 \
4 1   2 8 9 3 \
4 1   3 9 10 4 \
4 1   4 10 11 5 \
4 1   5 11 6 0 \
6 1   0 1 2 3 4 5 \
6 1   11 10 9 8 7 6 \
");

stringstream link2string("\
12 \
-14.605 4.1275 -3.73062 \
1.249372 3.01625 -3.73062 \
3.01625 1.249372 -3.73062 \
3.01625 -1.249372 -3.73062 \
1.249372 -3.01625 -3.73062 \
-14.605 -4.1275 -3.73062 \
-14.605 4.1275 3.73062 \
1.249372 3.01625 3.73062 \
3.01625 1.249372 3.73062 \
3.01625 -1.249372 3.73062 \
1.249372 -3.01625 3.73062 \
-14.605 -4.1275 3.73062 \
\
9 \
4 1   0 6 7 1 \
4 1   1 7 8 2 \
4 1   2 8 9 3 \
4 1   3 9 10 4 \
4 1   4 10 11 5 \
4 1   5 11 6 0 \
6 1   0 1 2 3 4 5 \
6 1   11 10 9 8 7 6 \
4 1   0 5 11 6 \
");

stringstream link3string("\
12 \
-15.08125 3.01625 -2.8575 \
0.7890768 1.905 -2.8575 \
1.905 0.7890768 -2.8575 \
1.905 -0.7890768 -2.8575 \
0.7890768 -1.905 -2.8575 \
-15.08125 -3.01625 -2.8575 \
-15.08125 3.01625 2.8575 \
0.7890768 1.905 2.8575 \
1.905 0.7890768 2.8575 \
1.905 -0.7890768 2.8575 \
0.7890768 -1.905 2.8575 \
-15.08125 -3.01625 2.8575 \
\
9 \
4 1   0 6 7 1 \
4 1   1 7 8 2 \
4 1   2 8 9 3 \
4 1   3 9 10 4 \
4 1   4 10 11 5 \
4 1   5 11 6 0 \
6 1   0 1 2 3 4 5 \
6 1   11 10 9 8 7 6 \
4 1   0 5 11 6 \
");

stringstream link4string("0 0");

stringstream link5string("\
8 \
-0.5 4.0 -8.5725 \
-0.5 -4.0 -8.5725 \
-0.5 -4.0 -7.0 \
-0.5 4.0 -7.0 \
0.5 4.0 -8.5725 \
0.5 -4.0 -8.5725 \
0.5 -4.0 -7.0 \
0.5 4.0 -7.0 \
\
6 \
4 1   0 4 5 1 \
4 1   1 5 6 2 \
4 1   2 6 7 3 \
4 1   3 7 4 0 \
4 1   0 1 2 3 \
4 1   7 6 5 4 \
");

// Gripper fingers
stringstream link6string("\
8 \
-0.5 1.0 -7.9 \
-0.5 0.0 -7.9 \
-0.5 0.0 0.0 \
-0.5 1.0 0.0 \
0.5 1.0 -7.9 \
0.5 0.0 -7.9 \
0.5 0.0 0.0 \
0.5 1.0 0.0 \
\
6 \
4 1   0 4 5 1 \
4 1   1 5 6 2 \
4 1   2 6 7 3 \
4 1   3 7 4 0 \
4 1   0 1 2 3 \
4 1   7 6 5 4 \
");

stringstream link7string("\
8 \
-0.5 0.0 -7.9 \
-0.5 -1.0 -7.9 \
-0.5 -1.0 0.0 \
-0.5 0.0 0.0 \
0.5 0.0 -7.9 \
0.5 -1.0 -7.9 \
0.5 -1.0 0.0 \
0.5 0.0 0.0 \
\
6 \
4 1   0 4 5 1 \
4 1   1 5 6 2 \
4 1   2 6 7 3 \
4 1   3 7 4 0 \
4 1   0 1 2 3 \
4 1   7 6 5 4 \
");


vector<Object> Robot::links = {Object(link0string), Object(link1string), Object(link2string), Object(link3string), Object(link4string), Object(link5string), Object(link6string), Object(link7string)};

const double Robot::d[5] = {27.2, 0., 0., 0., 10.5};
double Robot::theta[5];
const double Robot::a[5] = {0., 19.2, 19.2, 0., 0.};
const double Robot::alpha[5] = {deg2rad(90.), deg2rad(180.), 0., deg2rad(90.), 0.};
double Robot::thetaoffset[5] = {0., 0., 0., deg2rad(90.), 0.};
int Robot::stepsfromhome[5] = {0, 3060, 0, 0, 0};
double Robot::speed = 5;
int Robot::stepsize[5] = {100,100,100,100,100};
int Robot::nestloc[5];
bool Robot::exceededlimits = false;
bool Robot::gripperClosed = false;
Block * Robot::attachedblock = NULL;
Transform Robot::blockreltrans;


void World::checkinit(void) {
	if (!initcalled) { 
		cerr << endl << "ERROR: You must call \"init()\" before calling other robot functions!" << endl << endl;
		exit(1);
	}
}

void World::checkzero(void) {
	if (!zerocalled) { 
		cerr << endl << "ERROR: You must call \"zero()\" before calling nest() function!" << endl << endl;
		exit(1);
	}
}

void World::init() {
	const double q[4] = {1., 0., 0., 0.};
	cam.resetvpwidth(); 
	initcalled = true;

	blocks = new Block[::numblocks];
	ontop = new int[::numblocks];
	for (int i=0; i<::numblocks; i++) {					// Set up all the blocks in a stack
		blocks[i] = Block(GPoint(::numblocks + 3. - i, 1.5, 1), Transform(q, GPoint(23., 23., 2.*i)));
		ontop[i] = i+1;
	}
	ontop[::numblocks-1] = -1;			// top block has nothing on top; -1 means no block on top of this
}

void World::moverel(int stp0, int stp1, int stp2, int stp3, int stp4) {
	enum modes {run, next, step};		// Determines when we wait for keys
	static modes mode = run;

	checkinit();			// Prints error and exits if init not called yet

	int steps[5];		// working registers, initially copies of steps
	steps[0] = stp0;
	steps[1] = stp1;
	steps[2] = stp2;
	steps[3] = stp3;
	steps[4] = stp4;

	bool arrivedtarget = false;
	while (!arrivedtarget) {
		arrivedtarget = rob.moverel(steps, false) ;			// Note: moverel changes steps!!!!
		rob.checklimits();
		cam.drawbase();
		rob.putinfo(cam);			// Draw text on screen
		rob.draw(cam);
		drawblocks();
		rob.redrawfinger(cam);
		cam.refresh();

		int key;
		while (1) {
			int waittime = (mode==run || (mode==next && !arrivedtarget)) ? 300 : 0;
			if ((key = waitKey(waittime)) == 255) break;		// timeout
			if (key == 32 || key == 'p') key = waitKey(0);	// Pause then process the next key pushed
			if (key == 'q') {exit(1);}
			if (key == 'g') {mode = run; break;}
			else if (key == 'n') {mode = next; break;}
			else if (key == 's') {mode = step; break;}
			else if (key == 'l') {cam.shiftviewpoint(cam.camleft);}
			else if (key == 'r') {cam.shiftviewpoint(cam.camright);}
			else if (key == 'u') {cam.shiftviewpoint(cam.camup);}
			else if (key == 'd') {cam.shiftviewpoint(cam.camdown);}
			else if (key == 'i') {cam.shiftviewpoint(cam.camin);}
			else if (key == 'o') {cam.shiftviewpoint(cam.camout);}
			else break;		// Skip redraw if other keys
			cam.drawbase();
			rob.putinfo(cam);			// Draw text on screen
			rob.draw(cam);
			drawblocks();
			rob.redrawfinger(cam);
			cam.refresh();
		}
	} 
}

void World::drawblocks(int lineType) {
	if (blocks == NULL) return;

	for (int i=0; i<::numblocks; i++) {
//		blocks[i].draw(cam, BLUE, lineType);
		for (int thisblock = i; thisblock != -1; thisblock = ontop[thisblock]) {	// Draw this block and those on top
			blocks[thisblock].draw(cam, BLUE, lineType);
		}
	}
}

void World::nest(void) {
	checkzero();
	moverel(rob.getnestrelative());
}

void World::gripperOpen(void) { 
	rob.gripperOpen(blocks, ontop, ::numblocks); 
	moverel(0,0,0,0,0); 	// moverel will redraw
}

void World::gripperClose(void) { 
	rob.gripperClose(blocks, ontop, ::numblocks); 
	moverel(0,0,0,0,0); 
}

void World::shutdown(void) {
	waitKey(0);				// Wait for a keypress so you see final screen
}

void World::sleep(int ms) {
	waitKey(ms);			// Sleep by waiting for a button press and ignoring return key
}


static World theworld;

// Global functions
void init(void) {theworld.init();}

void moverel(int t0, int t1, int t2, int t3, int t4) {
	theworld.moverel(t0, t1, t2, t3, t4);
}

void zero(void) {
	theworld.zero();
}

void nest(void) {
	theworld.nest();
}

void gripperOpen(void) {
	theworld.gripperOpen();
}

void gripperClose(void) {
	theworld.gripperClose();
}

void shutdown(void) {
	theworld.shutdown();
}

void sleep(int ms) {
	theworld.sleep(ms);
}
