#ifndef __VLABVOLT_H
#define __VLABVOLT_H

#define PI 3.14159265358979323846
#define deg2rad(x) ((x) * PI/180.)
#define rad2deg(x) ((x) * 180./PI)

#include <vector>
#include <iostream>
#include <iomanip>
#include <fstream>
#include <opencv2/core.hpp>
#include <opencv2/imgproc.hpp>
#include <opencv2/highgui.hpp>

#include "virtuallabvolt.h"

#define RED   Scalar(0,0,255)
#define GREEN Scalar(0,255,0)
#define BLUE  Scalar(255,0,0)
#define MAGENTA Scalar(255,0,255)
#define CYAN  Scalar(255,255,0)
#define YELLOW Scalar(0,255,255)
#define BLACK  Scalar(0,0,0)
#define GRAY  Scalar(127,127,127)
#define WHITE  Scalar(255,255,255)

using namespace std;
using namespace cv;

class GPoint {           // Two- or Three-dimensional generlaized point
private:
  double X[3];
public:
//  GPoint(void)  {X[0] = X[1] = X[2] = 0.;}
	GPoint(double x, double y, double z = 0.0)  {X[0] = x; X[1] = y; X[2] = z;}
	GPoint(void) : GPoint(0., 0., 0.)  {}

	double operator[] (int i) const { return X[i]; }
	double& operator[] (int i) { return *(X+i); }

	GPoint& normalize(void);
	GPoint cross(const GPoint& p) const;              // return *this cross p
	double dot(const GPoint& p) const;

	GPoint& operator+= (const GPoint& p)  { X[0]+=p.X[0]; X[1]+=p.X[1]; X[2]+=p.X[2]; return *this; }
	GPoint& operator-= (const GPoint& p)  { X[0]-=p.X[0]; X[1]-=p.X[1]; X[2]-=p.X[2]; return *this; }
	GPoint& operator*= (double d)  { X[0]*=d; X[1]*=d; X[2]*=d; return *this; }
	GPoint& operator/= (double d)  { X[0]/=d; X[1]/=d; X[2]/=d; return *this; }
	GPoint operator+ (void) const  { return *this; }
	GPoint operator- (void) const  { return GPoint(-X[0], -X[1], -X[2]); }
	GPoint operator+ (const GPoint& p)  const { return GPoint(X[0]+p.X[0], X[1]+p.X[1], X[2]+p.X[2]); }
	GPoint operator- (const GPoint& p)  const { return GPoint(X[0]-p.X[0], X[1]-p.X[1], X[2]-p.X[2]); }
	GPoint operator* (double d) const  { return GPoint(X[0]*d, X[1]*d, X[2]*d); }
	GPoint operator/ (double d) const  { return GPoint(X[0]/d, X[1]/d, X[2]/d); }
//	int operator== (const GPoint& p) const
//    { return  (X==p.X) && (Y==p.Y) && (Z==p.Z); }
  friend istream& operator>> (istream& s, GPoint& p);
  friend ostream& operator<< (ostream& s, const GPoint& p);
};

inline GPoint operator*(double d, const GPoint& p) { return p*d; }


class Transform {
protected:
	double val[4][4];

public:
	Transform(void);
	Transform(const double q[4], const GPoint &t);
	Transform(double d, double theta, double a, double alpha);
	Transform(double rotZ, double rotV, double rotW, const GPoint &t);
	void setval(int i, int j, double v) { val[i][j] = v; }
	double getval(int i, int j) const { return val[i][j]; }
	Transform operator* (const Transform& t) const;
	Transform& operator*= (const Transform& t) {return *this = *this * t;} 
	GPoint operator* (const GPoint& p) const;
	GPoint rotatedPoint(const GPoint& p) const;
	GPoint gettranslation(void) const { return GPoint(val[0][3], val[1][3], val[2][3]); }
	void settranslation(const GPoint &g) { val[0][3] = g[0]; val[1][3] = g[1]; val[2][3] = g[2]; }
	Transform inverse(void) const; 		// return the inverse

	friend istream& operator>> (istream& s, Transform& p);
	friend ostream& operator<< (ostream& s, const Transform& p);
};


#define VIEWPORTWIDTH 500

class Viewport {
private:
	int width;
	const char* windowname;
	const Scalar backgroundcolor;
	Mat canvas;
public:
	Viewport(int w = VIEWPORTWIDTH, const char* wname = "labvoltview") ;
	~Viewport() { destroyWindow(windowname); } //destroy the created window

	int vpwidth(void) const { return width; }
	void resetwidth(void) { width = ::viewportwidth; refresh(); } 

	void refresh(void) ;

	void drawline(Point start, Point end, const Scalar & color = Scalar(0,0,0), int thickness = 1, int lineType = LINE_8) const ;
	inline void drawpoly(const Point * pts, int npts, const Scalar & color = WHITE, int lineType = LINE_8, int shift = 0) ;
	inline void drawpoly2(const Point * pts, int npts, const Scalar & color = WHITE, int lineType = LINE_8, int shift = 0) ; // Allows non-convex
	void drawpolyboundary(const Point * pts, int npts, const Scalar & color = BLACK, int thickness = 1, int lineType = LINE_8) const ;
	inline void puttext(const string& text, Point org = Point(5,15), int fontFace = FONT_HERSHEY_PLAIN, double fontScale = 1, 
		const Scalar & color = BLACK, int thickness=1, int lineType=8, bool bottomLeftOrigin=false ) ;
};

class Camera {
private:
	double scale;
	GPoint viewpoint;		// Position we are viewing from
	Transform vptrans; 		// Transform world wrt camera for this viewpoint -- keep synchronized
							// Note: not camera wrt world
	GPoint centershift;		// Shift of image in viewer
	Viewport viewport;
	GPoint lightsource;

	void setTransform(void);
	Point toImage(const GPoint& p) const ;		// Only x and y of result are meaningful

	Camera(const Camera &) ;

public:
	enum camdirection {camleft, camright, camup, camdown, camin, camout};
	Camera(double s, GPoint vp, GPoint cs = GPoint(-40,40)) : scale(s), viewpoint(vp), viewport(VIEWPORTWIDTH), centershift(cs),
              lightsource(GPoint(50,-200,400).normalize()) { setTransform(); }   
	void resetvpwidth(void) { viewport.resetwidth(); }
	void shiftviewpoint(camdirection direction) ; 
	const GPoint & getviewpoint(void) const { return viewpoint; }
	void drawbase(void) const ;
//													 LINE_8 is 8-connected line and LINE_4 is 4-connected line
	void drawline(GPoint start, GPoint end, const Scalar & color = BLACK, int thickness = 2, int lineType = LINE_8) const ;
	void drawpoly(const vector<GPoint> & gpts, GPoint surfnorm, const Scalar & color = WHITE, bool isconvex = 1, int lineType = LINE_8, int shift = 0) ;
	void puttext(const string& text, int linenum = 1, int fontFace = FONT_HERSHEY_PLAIN, double fontScale = 1, 
		const Scalar & color = BLACK, int thickness=1, int lineType=8, bool bottomLeftOrigin=false ) ;

	void refresh(void) { viewport.refresh(); }            // MAKE PRIVATE
};


class Frame : public Transform {
	public:
		Frame(const Transform & t = Transform()) : Transform(t) {}
		void draw(Camera & cam) const;
};


class Surface {
private:
	vector<GPoint> points;		// Points relative to a local (the link's) frame
	GPoint normal;				// normal to the surface
	bool isconvex;

	void setnormal(void) {normal = (points[1] - points[0]).cross(points[2] - points[1]).normalize();}

public:
	Surface(void) : isconvex(true) { }	

	void resurface(const vector<GPoint> & pts, const vector<int> & indices, bool iscvx) ;	// only called by Object constructor
	void resurface4(const GPoint & p0, const GPoint & p1, const GPoint & p2, const GPoint & p3)  ;
	void draw(Camera & cam, Transform t, const Scalar & color = WHITE, int lineType = LINE_8) const ;
};


class Object {
protected:
	vector<Surface> surfaces;
	Transform trans;

	Object(Transform t) : surfaces(6), trans(t) { }		// Used by Block constructor

public:
	Object(void) { }						// Needed by block constructor for array of Blocks
	Object(stringstream & strobj) ;

	void setTransform(const Transform & t) {trans = t;}
	const Transform& getTransform(void) const {return trans;}
	GPoint gettranslation(void) const { return trans.gettranslation(); }

	void draw(Camera & cam, const Scalar & color = WHITE, int lineType = LINE_8) const ;
};

class Block : public Object {
private:
	GPoint dims2;		// dimensions divided by 2 -- note that origin is in the center

public:
	Block(void) : Object() {}
	Block(GPoint d2, Transform t);
	const GPoint& getdims2(void) const { return dims2; }
	void draw(Camera & cam, const Scalar & color = GREEN, int lineType = LINE_8) const {Object::draw(cam, color, lineType);}
};


class Robot {
private:
	static vector<Object> links;
	static const double d[5];			// Joint parameters
	static double theta[5];
	static const double a[5];
	static const double alpha[5];
	static double thetaoffset[5];		// Account for +90 in joint 4
	static int stepsfromhome[5];
	static double speed;
	static int stepsize[5];
	static int nestloc[5];
	static bool exceededlimits;
	static bool gripperClosed;
	static Block * attachedblock;
	static Transform blockreltrans;
	
	void convertsteps(void);
	void setTransforms(void);
public:
	Robot(void) {convertsteps(); setTransforms();}

//	Object& getLink(int i) {return links[i];}                // for testing
	void setTransformZero(const Transform& t) { links[0].setTransform(t); setTransforms(); }
	void draw(Camera & cam, int lineType = LINE_8) ;
	void redrawfinger(Camera & cam, int lineType = LINE_8) ;	// Redraw finger which may have been covered by a block
	bool moverel(int steps[], bool fullmove) ;		// Return true if the move was completed
	void checklimits(void) ;			// Return true if all axes within limits and z above table
	void putinfo(Camera & cam) ;				// put information on screen
	void zero(void) { for (int i=0; i<5; i++) nestloc[i] = stepsfromhome[i]; }
	const vector<int> getnestrelative(void) ; 		// return moverel steps to get to nest
	void gripperOpen(Block * blocks, int *ontop, int nblocks) ;
	void gripperClose(Block * blocks, int *ontop, int nblocks) ;
};


class World {
private:
	Robot rob;
	Camera cam;
	Block *blocks;		// Set up on call to init() as we don't know the number until run time
	int *ontop;			// Index of a block on top of this block

	bool initcalled;
	void checkinit(void) ;
	bool zerocalled;
	void checkzero(void) ;
public:
	World(void) : cam(10, GPoint(400, -400, 99)), blocks(NULL), ontop(NULL), initcalled(false), zerocalled(false) { }
	~World(void) { delete[] blocks; delete[] ontop;}

	void init() ;
	void moverel(int stp0, int stp1, int stp2, int stp3, int stp4) ;
	void moverel(const vector<int> steps) { moverel(steps[0], steps[1], steps[2], steps[3], steps[4]); } 	// maybe not used
//	void move(void);
	void zero(void) { rob.zero(); zerocalled = true; }
	void drawblocks(int lineType = LINE_8) ;
	void nest(void);
	void gripperOpen(void) ;
	void gripperClose(void) ;
	void shutdown(void);
	void sleep(int ms);			// Sleep function if needed

};

#if 0
// Global functions

void init(void);
void moverel(int t0, int t1, int t2, int t3, int t4) ;
void zero(void);
void nest(void);
void gripperOpen(void);
void gripperClose(void);
void shutdown(void);
void sleep(int ms);			// Sleep function if needed
#endif
	
#endif
